@@ -50,6 +50,10 @@
 #include <xf86platformBus.h>
 #endif
 
+#ifndef XF86_ALLOCATE_GPU_SCREEN
+#define XF86_ALLOCATE_GPU_SCREEN 0
+#endif
+
 static const struct intel_device_info intel_generic_info = {
 	.gen = -1,
 };
@@ -224,6 +228,38 @@ static const SymTabRec intel_chipsets[] = {
 	{0x0155, "HD Graphics"},
 	{0x0157, "HD Graphics"},
 
+	/* Broadwell Marketing names */
+	{0x1602, "HD graphics"},
+	{0x1606, "HD graphics"},
+	{0x160B, "HD graphics"},
+	{0x160A, "HD graphics"},
+	{0x160D, "HD graphics"},
+	{0x160E, "HD graphics"},
+	{0x1612, "HD graphics 5600"},
+	{0x1616, "HD graphics 5500"},
+	{0x161B, "HD graphics"},
+	{0x161A, "HD graphics"},
+	{0x161D, "HD graphics"},
+	{0x161E, "HD graphics 5300"},
+	{0x1622, "Iris Pro graphics 6200"},
+	{0x1626, "HD graphics 6000"},
+	{0x162B, "Iris graphics 6100"},
+	{0x162A, "Iris Pro graphics P6300"},
+	{0x162D, "HD graphics"},
+	{0x162E, "HD graphics"},
+	{0x1632, "HD graphics"},
+	{0x1636, "HD graphics"},
+	{0x163B, "HD graphics"},
+	{0x163A, "HD graphics"},
+	{0x163D, "HD graphics"},
+	{0x163E, "HD graphics"},
+
+	/* When adding new identifiers, also update:
+	 * 1. intel_identify()
+	 * 2. man/intel.man
+	 * 3. README
+	 */
+
 	{-1, NULL} /* Sentinel */
 };
 
@@ -237,7 +273,7 @@ static const struct pci_id_match intel_device_match[] = {
 
 #if KMS
 	INTEL_I830_IDS(&intel_i830_info),
-	INTEL_I845G_IDS(&intel_i830_info),
+	INTEL_I845G_IDS(&intel_i845_info),
 	INTEL_I85X_IDS(&intel_i855_info),
 	INTEL_I865G_IDS(&intel_i865_info),
 
@@ -401,9 +437,9 @@ static void intel_identify(int flags)
 	if (unique != stack)
 		free(unique);
 
-	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) HD Graphics: 2000-5000\n");
-	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) Iris(TM) Graphics: 5100\n");
-	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) Iris(TM) Pro Graphics: 5200\n");
+	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) HD Graphics: 2000-6000\n");
+	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) Iris(TM) Graphics: 5100, 6100\n");
+	xf86Msg(X_INFO, INTEL_NAME ": Driver for Intel(R) Iris(TM) Pro Graphics: 5200, 6200, P6300\n");
 }
 
 static Bool intel_driver_func(ScrnInfoPtr pScrn,
@@ -453,7 +489,7 @@ _xf86findDriver(const char *ident, XF86ConfDevicePtr p)
 	return NULL;
 }
 
-static enum accel_method { SNA, UXA, GLAMOR } get_accel_method(void)
+static enum accel_method { NOACCEL, SNA, UXA, GLAMOR } get_accel_method(void)
 {
 	enum accel_method accel_method = DEFAULT_ACCEL_METHOD;
 	XF86ConfDevicePtr dev;
@@ -467,7 +503,9 @@ static enum accel_method { SNA, UXA, GLAMOR } get_accel_method(void)
 
 		s = xf86FindOptionValue(dev->dev_option_lst, "AccelMethod");
 		if (s ) {
-			if (strcasecmp(s, "sna") == 0)
+			if (strcasecmp(s, "none") == 0)
+				accel_method = NOACCEL;
+			else if (strcasecmp(s, "sna") == 0)
 				accel_method = SNA;
 			else if (strcasecmp(s, "uxa") == 0)
 				accel_method = UXA;
@@ -511,7 +549,7 @@ intel_scrn_create(DriverPtr		driver,
 	scrn->driverVersion = INTEL_VERSION;
 	scrn->driverName = (char *)INTEL_DRIVER_NAME;
 	scrn->name = (char *)INTEL_NAME;
-	scrn->driverPrivate = (void *)(match_data | 1);
+	scrn->driverPrivate = (void *)(match_data | (flags & XF86_ALLOCATE_GPU_SCREEN) | 2);
 	scrn->Probe = NULL;
 
 	if (xf86IsEntitySharable(entity_num))
@@ -526,9 +564,14 @@ intel_scrn_create(DriverPtr		driver,
 #if KMS
 	switch (get_accel_method()) {
 #if USE_SNA
-	case SNA: return sna_init_scrn(scrn, entity_num);
+	case NOACCEL:
+	case SNA:
+		return sna_init_scrn(scrn, entity_num);
 #endif
 #if USE_UXA
+#if !USE_SNA
+	case NOACCEL:
+#endif
 	case GLAMOR:
 	case UXA:
 		  return intel_init_scrn(scrn);
